//
// Created by philipp on 30.11.17.
//

#include <boost/asio/ip/tcp.hpp>
#include <iostream>
#include "../include/PlatoonProtocolLib/FollowerVehicle.h"
#include "../include/PlatoonProtocolLib/Utils.h"

using namespace platoonProtocol;
using namespace std::chrono_literals;

constexpr networking::time::Duration FollowerVehicle::HEARTBEAT_INTERVAL;
constexpr networking::time::Duration FollowerVehicle::BROADCAST_TIMEOUT;

FollowerVehicle::Ptr FollowerVehicle::create(networking::Networking & net, VehicleId ownVehicleId)
{
    return std::make_shared<FollowerVehicle>(PrivateTag{}, net, ownVehicleId);
}

FollowerVehicle::FollowerVehicle(PrivateTag, networking::Networking & net, VehicleId ownVehicleId)
    : Vehicle(net, ownVehicleId, Role::FOLLOWER)
{
    heartbeatTimer = networking::time::Timer::create(net);
    heartbeatSender = networking::message::DatagramSender<PlatoonMessage>::create(net);
    broadcastReceiver = networking::message::DatagramReceiver<PlatoonMessage>::create(
        net, getOwnEndpoint().getUdpPort(), PlatoonMessage::MAX_SIZE);
}

void FollowerVehicle::stop()
{
    Vehicle::stop();

    stopHeartbeat();

    stopReceiveBroadcast();
}

networking::time::TimedValue<PlatoonConfig> FollowerVehicle::getPlatoonConfig()
{
    return platoonConfig.getNonAtomicCopy();
}

void FollowerVehicle::doCreatePlatoon()
{
    if (state != State::IDLE)
        return;

    setState(State::CREATING_PLATOON);

    // Reset leader.
    leader = VehicleEndpoint{};

    Vehicle::doCreatePlatoon();
}

void FollowerVehicle::doRunPlatoon()
{
    if (state != State::CREATING_PLATOON)
        return;

    setState(State::RUNNING_PLATOON);

    platoonCreateServiceServer->stop();

    startHeartbeat();

    startReceiveBroadcast();

    onRunningPlatoonCallback();
}

void FollowerVehicle::doLeavePlatoon()
{
    stop();

    if (state == State::RUNNING_PLATOON)
    {
        sendLeavePlatoonMessage();
        onLeavingPlatoonCallback();
    }

    setState(State::IDLE);
}

void FollowerVehicle::makePlatoonCreateRequest(const VehicleEndpoint & vehicleEndpoint)
{
    // We'll send a request indicating that we want to be the follower.
    auto self = shared_from_this();
    newPlatoonCreateServiceClient();
    platoonCreateServiceClient->asyncCall(
        PlatoonMessage::followerRequest(getOwnEndpoint().getVehicleId()),
        vehicleEndpoint.getHost(),
        vehicleEndpoint.getTcpPort(),
        3s,
        [self, vehicleEndpoint](const auto & error, const auto & response)
        {
            if (self->state != State::CREATING_PLATOON)
                return;

            if (error)
                return;

            if (response.vehicleId != vehicleEndpoint.getVehicleId() ||
                response.messageType != messageTypes::ACCEPT_RESPONSE)
                return;

            self->leader = vehicleEndpoint;
            self->platoonId = response.platoonId;
            self->doRunPlatoon();
        });
}

void FollowerVehicle::makePlatoonCreateResponse(const networking::Resolver::Endpoint & clientEndpoint,
                                                const PlatoonMessage & request,
                                                PlatoonMessage & response)
{
    if (state != State::CREATING_PLATOON)
    {
        response = PlatoonMessage::rejectResponse(getOwnEndpoint().getVehicleId());
        return;
    }

    auto vehicleId = request.vehicleId;
    // Only consider this message if it is a leader request,
    // meaning that a leader wants us to join its platoon.
    if (request.messageType != messageTypes::LV_REQUEST || vehicleId == 0)
    {
        response = PlatoonMessage::rejectResponse(getOwnEndpoint().getVehicleId());
        return;
    }

    leader = VehicleEndpoint{request.vehicleId, clientEndpoint.ip};
    platoonId = request.platoonId;

    doRunPlatoon();

    response = PlatoonMessage::acceptResponse(getOwnEndpoint().getVehicleId(), platoonId);
}

void FollowerVehicle::startHeartbeat()
{
    auto self = shared_from_this();
    heartbeatTimer = networking::time::Timer::create(net);
    heartbeatTimer->startPeriodicTimeout(
        HEARTBEAT_INTERVAL,
        [self]
        {
            self->heartbeatSender->asyncSend(
                PlatoonMessage::heartbeatMessage(self->getOwnEndpoint().getVehicleId(), self->platoonId),
                self->leader.getHost(),
                self->leader.getUdpPort(),
                HEARTBEAT_INTERVAL,
                [self](auto && ...)
                {});
        });
}

void FollowerVehicle::stopHeartbeat()
{
    heartbeatTimer->stop();
    heartbeatSender->stop();
}

void FollowerVehicle::startReceiveBroadcast()
{
    lastBroadcastTimestamp = networking::time::now();
    receiveBroadcast();
}

void FollowerVehicle::receiveBroadcast()
{
    auto self = shared_from_this();
    broadcastReceiver->asyncReceive(
        BROADCAST_TIMEOUT,
        [self](const auto & error, const PlatoonMessage & message, const auto & host, auto port)
        {
            if (self->state != State::RUNNING_PLATOON)
                return;

            auto timeSinceLastBroadcast = networking::time::now() - self->lastBroadcastTimestamp;

            // Broadcast-Timeout or manually stopped.
            if (error == networking::error::codes::ABORTED ||
                timeSinceLastBroadcast >= BROADCAST_TIMEOUT)
            {
                self->doLeavePlatoon();
                return;
            }
            else if (error ||
                     message.vehicleId != self->leader.getVehicleId() ||
                     message.platoonId != self->platoonId)
            {
                // In case of any other error we just ignore the message and receive again.
                self->receiveBroadcast();
                return;
            }

            if (message.messageType == messageTypes::LV_BROADCAST)
            {
                PlatoonConfig newConfig{message.platoonSpeed, message.innerPlatoonDistance};
                if (self->platoonConfig.get() != newConfig)
                {
                    self->platoonConfig = newConfig;
                    self->onPlatoonConfigUpdatedCallback();
                }
                // Clear the broadcast-timeout.
                self->lastBroadcastTimestamp = networking::time::now();
            }

            // Receive the next broadcast.
            self->receiveBroadcast();
        });
}

void FollowerVehicle::stopReceiveBroadcast()
{
    broadcastReceiver->stop();
}

void FollowerVehicle::sendLeavePlatoonMessage()
{
    auto self = shared_from_this();
    heartbeatSender->asyncSend(
        PlatoonMessage::leaveMessage(getOwnEndpoint().getVehicleId(), platoonId),
        leader.getHost(),
        leader.getUdpPort(),
        HEARTBEAT_INTERVAL,
        [self](auto && ...)
        {});
}

FollowerVehicle::Ptr FollowerVehicle::shared_from_this()
{
    return std::static_pointer_cast<FollowerVehicle>(Vehicle::shared_from_this());
}
